/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

//* TODO pick two points: show locations and distance

#define FE_RUL_DEBUG		FALSE

#include "operator/operator.pmh"

namespace fe
{

void RulerOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<bool>("orthoGrid")=true;
	catalog<String>("orthoGrid","label")="Ortho Grid";

	catalog<bool>("behind")=true;
	catalog<String>("behind","label")="Behind Object";
	catalog<String>("behind","enabler")="orthoGrid";

	catalog<bool>("crossGrid")=false;
	catalog<String>("crossGrid","label")="Cross Grid";

	catalog<bool>("outline")=false;
	catalog<String>("outline","label")="Outline";

	catalog<Real>("base")=10.0;
	catalog<Real>("base","min")=2.0;
	catalog<Real>("base","max")=10.0;
	catalog<String>("base","label")="Base";

	catalog<Real>("excess")=0.0;
	catalog<Real>("excess","high")=10.0;
	catalog<Real>("excess","max")=100.0;
	catalog<String>("excess","label")="Excess";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","visible")="any";
	catalog<String>("Brush","prompt")="Prompt here.";

	catalog< sp<Component> >("Input Surface");

	m_spBrush=new DrawMode();
	m_spBrush->setDrawStyle(DrawMode::e_foreshadow);
	m_spBrush->setLineWidth(2.0);
	m_spBrush->setAntialias(TRUE);
	m_spBrush->setLit(FALSE);

	m_spHaze=new DrawMode();
	m_spHaze->setDrawStyle(DrawMode::e_solid);
	m_spHaze->setLineWidth(2.0);
	m_spHaze->setAntialias(TRUE);
	m_spHaze->setBackfaceCulling(FALSE);
	m_spHaze->setLit(FALSE);

	m_spOverlay=new DrawMode();
	m_spOverlay->setDrawStyle(DrawMode::e_solid);
	m_spOverlay->setLit(FALSE);
}

void RulerOp::handle(Record& a_rSignal)
{
#if FE_RUL_DEBUG
	feLog("RulerOp::handle\n");
#endif

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

	if(spDrawBrush.isValid() && m_spInputAccessible.isValid())
	{
		const BWORD orthoGrid=catalog<bool>("orthoGrid");
		const BWORD behind=catalog<bool>("behind");
		const BWORD crossGrid=catalog<bool>("crossGrid");
		const BWORD outline=catalog<bool>("outline");

		const Real excess=catalog<Real>("excess");
		const Real base=catalog<Real>("base");

		const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
		const SpatialVector& rRayDirection=rayDirection(a_rSignal);

		sp<ViewI> spView=spDrawOverlay->view();
		sp<ViewMode> spViewMode=spView->viewMode();
		const Box2i viewport=spViewMode->viewport();

		const I32 width=viewport.size()[0];
		const I32 height=viewport.size()[1];

		SpatialTransform camera=spViewMode->camera();
		SpatialTransform invCamera;
		invert(invCamera,camera);

		const SpatialVector cameraDir= -invCamera.column(2);
		const SpatialVector cameraPos=invCamera.column(3);

		m_event.bind(windowEvent(a_rSignal));

#if FE_RUL_DEBUG
		feLog("%s\n",c_print(m_event));
#endif


#if FE_RUL_DEBUG
		feLog("RulerOp::handle brush done\n");
#endif

		sp<SurfaceAccessorI> spInputPosition;
		if(!access(spInputPosition,m_spInputAccessible,
				e_point,e_position)) return;

		const U32 pointCount=spInputPosition->count();
		if(!pointCount)
		{
			catalog<String>("warning")+="No input points;";
			return;
		}

		sp<SurfaceAccessorI> spInputVertices;
		if(!access(spInputVertices,m_spInputAccessible,
				e_primitive,e_vertices)) return;

		spDrawBrush->pushDrawMode(m_spBrush);
		spDrawOverlay->pushDrawMode(m_spOverlay);

		SpatialVector minCorner;
		SpatialVector maxCorner;
		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point=
					spInputPosition->spatialVector(pointIndex);
			for(U32 m=0;m<3;m++)
			{
				if(!pointIndex || minCorner[m]>point[m])
				{
					minCorner[m]=point[m];
				}
				if(!pointIndex || maxCorner[m]<point[m])
				{
					maxCorner[m]=point[m];
				}
			}
		}

		SpatialVector boundSize=maxCorner-minCorner;
		SpatialVector center=minCorner+0.5*boundSize;

		const I32 pixels=256;
		const Real worldSize=spView->worldSize(center,pixels);

		//* log_n(x) = log_d(x)/log_d(b)  (for any convenient d)
		const I32 intLog=(worldSize>0.0)? (I32)(log(worldSize)/log(base)): 0;
		const Real step=pow(base,intLog);

		const Real foreAlpha=0.3;
		const Color hardBlack(0.0,0.0,0.0);
		const Color alphaBlack(0.0,0.0,0.0,foreAlpha);
		const Color grey(0.6,0.6,0.6);
		const Color white(1.0,1.0,1.0);
		const Color red(1.0,0.0,0.0,foreAlpha);
		const Color green(0.0,1.0,0.0,foreAlpha);
		const Color blue(0.0,0.0,1.0,foreAlpha);
		const Color cyan(0.0,1.0,1.0,foreAlpha);
		const Color magenta(1.0,0.0,1.0,foreAlpha);
		const Color yellow(1.0,1.0,0.0,foreAlpha);

		const Real hazeAlpha=0.1;
		const Color hazeCyan(0.0,1.0,1.0,hazeAlpha);
		const Color hazeMagenta(1.0,0.0,1.0,hazeAlpha);
		const Color hazeYellow(1.0,1.0,0.0,hazeAlpha);

		const Real beyond=0.01*step;

		const Real startX=floor((minCorner[0]-excess)/step)*step;
		const Real endX=floor((maxCorner[0]+excess)/step+1.0)*step+beyond;

		const Real startY=floor((minCorner[1]-excess)/step)*step;
		const Real endY=floor((maxCorner[1]+excess)/step+1.0)*step+beyond;

		const Real startZ=floor((minCorner[2]-excess)/step)*step;
		const Real endZ=floor((maxCorner[2]+excess)/step+1.0)*step+beyond;

		const BWORD highX=((cameraDir[0]>0.0) == behind);
		const BWORD highY=((cameraDir[1]>0.0) == behind);
		const BWORD highZ=((cameraDir[2]>0.0) == behind);

		const Real backX=highX? endX: startX;
		const Real backY=highY? endY: startY;
		const Real backZ=highZ? endZ: startZ;

		const SpatialVector backM(backX,backY,backZ);

#if FALSE	//FE_RUL_DEBUG
		feLog("RulerOp::handle bounds min %s max %s\n",
				c_print(minCorner),c_print(maxCorner));
		feLog("  worldSize %.6G intLog %d step %.6G\n",
				worldSize,intLog,step);
		feLog("  x %.6G to %.6G back %.6G\n",startX,endX,backX);
		feLog("  y %.6G to %.6G back %.6G\n",startY,endY,backY);
		feLog("  z %.6G to %.6G back %.6G\n",startZ,endZ,backZ);
#endif

		SpatialVector planeCenter(0.0,0.0,0.0);
		SpatialVector planeNormal(0.0,1.0,0.0);

		SpatialVector line[2];
		const U32 margin=1;
		String text;

		if(orthoGrid)
		{
			spDrawBrush->pushDrawMode(m_spHaze);

			SpatialVector wallVert[4];
			SpatialVector wallNorm[4];

			set(wallVert[0],backX,startY,startZ);
			set(wallVert[1],backX,startY,endZ);
			set(wallVert[2],backX,endY,startZ);
			set(wallVert[3],backX,endY,endZ);

			set(wallNorm[0],1.0,0.0,0.0);
			set(wallNorm[1],1.0,0.0,0.0);
			set(wallNorm[2],1.0,0.0,0.0);
			set(wallNorm[3],1.0,0.0,0.0);

			spDrawBrush->drawTriangles(wallVert,wallNorm,NULL,4,
					DrawI::e_strip,FALSE,&hazeCyan);

			set(wallVert[0],startX,backY,startZ);
			set(wallVert[1],startX,backY,endZ);
			set(wallVert[2],endX,backY,startZ);
			set(wallVert[3],endX,backY,endZ);

			set(wallNorm[0],0.0,1.0,0.0);
			set(wallNorm[1],0.0,1.0,0.0);
			set(wallNorm[2],0.0,1.0,0.0);
			set(wallNorm[3],0.0,1.0,0.0);

			spDrawBrush->drawTriangles(wallVert,wallNorm,NULL,4,
					DrawI::e_strip,FALSE,&hazeMagenta);

			set(wallVert[0],startX,startY,backZ);
			set(wallVert[1],startX,endY,backZ);
			set(wallVert[2],endX,startY,backZ);
			set(wallVert[3],endX,endY,backZ);

			set(wallNorm[0],0.0,0.0,1.0);
			set(wallNorm[1],0.0,0.0,1.0);
			set(wallNorm[2],0.0,0.0,1.0);
			set(wallNorm[3],0.0,0.0,1.0);

			spDrawBrush->drawTriangles(wallVert,wallNorm,NULL,4,
					DrawI::e_strip,FALSE,&hazeYellow);

			spDrawBrush->popDrawMode();

			const Real stepShort=0.1*step;

			//* TODO loop through X,Y,Z

			for(Real gridX=startX;gridX<endX;gridX+=step)
			{
				const BWORD onEdge=(fabs(gridX-backX)<stepShort);

				set(line[0],gridX,startY,backZ);
				set(line[1],gridX,endY,backZ);

				spDrawBrush->drawLines(line,NULL,2,
						DrawI::e_discrete,false,onEdge? &green: &yellow);

				if(!onEdge)
				{
					set(line[0],gridX,backY,startZ);
					set(line[1],gridX,backY,endZ);

					spDrawBrush->drawLines(line,NULL,
							2,DrawI::e_discrete,false,&magenta);

					text.sPrintf("%.6G",gridX);

					const SpatialVector edgePoint(gridX,backY,backZ);

					const Vector2i projected=spView->project(
							edgePoint,ViewMode::e_perspective);

					drawLabel(spDrawOverlay,projected,text,
							FALSE,margin,white,NULL,&blue);
				}

				if(crossGrid)
				{
					set(planeCenter,gridX,0.0,0.0);
					set(planeNormal,1.0,0.0,0.0);
					drawCrossSection(spDrawBrush,spInputVertices,
							planeCenter,planeNormal,cyan);
				}
			}

			for(Real gridY=startY;gridY<endY;gridY+=step)
			{
				const BWORD onEdge=(fabs(gridY-backY)<stepShort);

				set(line[0],backX,gridY,startZ);
				set(line[1],backX,gridY,endZ);

				spDrawBrush->drawLines(line,NULL,2,
						DrawI::e_discrete,false,onEdge? &blue: &cyan);

				if(!onEdge)
				{
					set(line[0],startX,gridY,backZ);
					set(line[1],endX,gridY,backZ);

					spDrawBrush->drawLines(line,NULL,2,
							DrawI::e_discrete,false,&yellow);

					text.sPrintf("%.6G",gridY);

					const SpatialVector edgePoint(backX,gridY,backZ);

					const Vector2i projected=spView->project(
							edgePoint,ViewMode::e_perspective);

					drawLabel(spDrawOverlay,projected,text,
							FALSE,margin,white,NULL,&blue);
				}

				if(crossGrid)
				{
					set(planeCenter,0.0,gridY,0.0);
					set(planeNormal,0.0,1.0,0.0);
					drawCrossSection(spDrawBrush,spInputVertices,
							planeCenter,planeNormal,magenta);
				}
			}

			for(Real gridZ=startZ;gridZ<endZ;gridZ+=step)
			{
				const BWORD onEdge=(fabs(gridZ-backZ)<stepShort);

				set(line[0],startX,backY,gridZ);
				set(line[1],endX,backY,gridZ);

				spDrawBrush->drawLines(line,NULL,2,
						DrawI::e_discrete,false,onEdge? &red: &magenta);

				if(!onEdge)
				{
					set(line[0],backX,startY,gridZ);
					set(line[1],backX,endY,gridZ);

					spDrawBrush->drawLines(line,NULL,2,
							DrawI::e_discrete,false,&cyan);

					text.sPrintf("%.6G",gridZ);

					const SpatialVector edgePoint(backX,backY,gridZ);

					const Vector2i projected=spView->project(
							edgePoint,ViewMode::e_perspective);

					drawLabel(spDrawOverlay,projected,text,
							FALSE,margin,white,NULL,&blue);
				}

				if(crossGrid)
				{
					set(planeCenter,0.0,0.0,gridZ);
					set(planeNormal,0.0,0.0,1.0);
					drawCrossSection(spDrawBrush,spInputVertices,
							planeCenter,planeNormal,yellow);
				}
			}

			SpatialVector pointArray[pointCount];
			Color colorArray[pointCount];

			for(U32 m=0;m<3;m++)
			{
				for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const SpatialVector point=
							spInputPosition->spatialVector(pointIndex);

					pointArray[pointIndex]=point;
					colorArray[pointIndex]=hardBlack;

					pointArray[pointIndex][m]=backM[m];
					colorArray[pointIndex][m]=
							(point[m]-minCorner[m])/boundSize[m];
				}

				spDrawBrush->drawPoints(pointArray,NULL,
						pointCount,true,colorArray);
			}
		}

		Vector2i location(0.5*width,height-16);
		const BWORD centered=TRUE;
		const Real dotRadius=3.0;
		const I32 yInc=12;

		BWORD impacted=FALSE;
		SpatialVector intersection(0.0,0.0,0.0);
		SpatialVector norm(0.0,1.0,0.0);
		I32 triIndex= -1;
		SpatialBary barycenter(0.0,0.0);

		if(m_spInput.isValid())
		{
			const Real maxDistance= -1.0;
			sp<SurfaceI::ImpactI> spImpact=m_spInput->rayImpact(
					rRayOrigin,rRayDirection,maxDistance);
			if(spImpact.isValid())
			{
				impacted=TRUE;
				intersection=spImpact->intersection();
				norm=spImpact->normal();
				barycenter=spImpact->barycenter();

				sp<SurfaceSearchable::Impact> spSearchImpact=spImpact;
				if(spSearchImpact.isValid())
				{
					triIndex=spSearchImpact->triangleIndex();
				}

				line[0]=intersection;

				for(U32 m=0;m<3;m++)
				{
					line[1]=line[0];
					line[1][m]=backM[m];

					Color axisColor=hardBlack;
					axisColor[m]=1.0;
					axisColor[3]=foreAlpha;

					spDrawBrush->drawLines(line,NULL,2,
							DrawI::e_discrete,false,&axisColor);
				}

				drawDot(spView,spDrawOverlay,line[0],dotRadius,white);
			}
		}

		if(m_event.isKeyPress(WindowEvent::e_keyEscape))
		{
			m_pickedA=FALSE;
			m_pickedB=FALSE;
		}

		if(m_event.isMouseRelease(WindowEvent::e_itemLeft))
		{
			if(impacted)
			{
				m_pointA=intersection;
				m_normA=norm;
				m_triIndexA=triIndex;
				m_barycenterA=barycenter;
				m_pickedA=TRUE;
			}
			else
			{
				m_pickedA=FALSE;
			}
		}

		if(m_event.isMouseRelease(WindowEvent::e_itemMiddle))
		{
			if(impacted)
			{
				m_pointB=intersection;
				m_normB=norm;
				m_triIndexB=triIndex;
				m_barycenterB=barycenter;
				m_pickedB=TRUE;
			}
			else
			{
				m_pickedB=FALSE;
			}
		}

		planeCenter=intersection;
		planeNormal=norm;
		BWORD triPlane=TRUE;
		BWORD showCrossSection=impacted;

		if(m_pickedA)
		{
			if(m_triIndexA>=0 && m_spInput.isValid())
			{
				sp<SurfaceI::ImpactI> spImpact=m_spInput->sampleImpact(
						m_triIndexA,m_barycenterA);
				m_pointA=spImpact->intersection();
				m_normA=spImpact->normal();
			}

			drawDot(spView,spDrawOverlay,m_pointA,dotRadius,hardBlack);

			text.sPrintf("A %s",c_print(m_pointA));

			drawLabel(spDrawOverlay,location,text,
					centered,margin,white,NULL,&blue);
			location[1]-=yInc;

			if(!m_pickedB)
			{
				planeCenter=m_pointA;
				showCrossSection=TRUE;
			}
		}

		if(m_pickedB)
		{
			if(m_triIndexB>=0 && m_spInput.isValid())
			{
				sp<SurfaceI::ImpactI> spImpact=m_spInput->sampleImpact(
						m_triIndexB,m_barycenterB);
				m_pointB=spImpact->intersection();
				m_normB=spImpact->normal();
			}

			drawDot(spView,spDrawOverlay,m_pointB,dotRadius,grey);

			text.sPrintf("B %s",c_print(m_pointB));

			drawLabel(spDrawOverlay,location,text,
					centered,margin,white,NULL,&blue);
			location[1]-=yInc;

			if(!m_pickedA)
			{
				const SpatialVector away=intersection-m_pointB;
				const Real awayDot=dot(away,m_normB);
				planeCenter=m_pointB+awayDot*m_normB;
				planeNormal=m_normB;
				triPlane=FALSE;
			}
		}

		if(m_pickedA && m_pickedB)
		{
			const SpatialVector delta=m_pointB-m_pointA;
			const Real distance=magnitude(delta);

			text.sPrintf("delta %s",c_print(delta));

			drawLabel(spDrawOverlay,location,text,
					centered,margin,white,NULL,&blue);
			location[1]-=yInc;

			text.sPrintf("distance %.6G",distance);

			drawLabel(spDrawOverlay,location,text,
					centered,margin,white,NULL,&blue);
			location[1]-=yInc;

			planeNormal=unitSafe(cross(delta,m_normA+m_normB));
			triPlane=FALSE;
		}

		if(impacted)
		{
			text.sPrintf("%s",c_print(intersection));

			drawLabel(spDrawOverlay,location,text,
					centered,margin,white,NULL,&blue);
			location[1]-=yInc;
		}

		if(showCrossSection)
		{
			if(triPlane)
			{
				const SpatialVector origCenter=planeCenter;

				set(planeCenter,origCenter[0],0.0,0.0);
				set(planeNormal,1.0,0.0,0.0);
				drawCrossSection(spDrawBrush,spInputVertices,
						planeCenter,planeNormal,cyan);

				set(planeCenter,0.0,origCenter[1],0.0);
				set(planeNormal,0.0,1.0,0.0);
				drawCrossSection(spDrawBrush,spInputVertices,
						planeCenter,planeNormal,magenta);

				set(planeCenter,0.0,0.0,origCenter[2]);
				set(planeNormal,0.0,0.0,1.0);
				drawCrossSection(spDrawBrush,spInputVertices,
						planeCenter,planeNormal,yellow);
			}
			else
			{
				drawCrossSection(spDrawBrush,spInputVertices,
					planeCenter,planeNormal,yellow);
			}
		}

		if(outline)
		{
			//* NOTE need face normals for sharp facets

			sp<SurfaceAccessorI> spInputNormals;
			if(!access(spInputNormals,m_spInputAccessible,
					e_primitive,e_normal,e_quiet))
			{
				catalog<String>("message")+="generated face normals;";
			}

//			feLog("draw outline\n");

			std::map<U64,Edge> edgeMap;

			const U32 primitiveCount=spInputVertices->count();

			for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
					primitiveIndex++)
			{
				if(interrupted())
				{
					break;
				}

				const U32 subCount=
						spInputVertices->subCount(primitiveIndex);

				SpatialVector faceNormal(0.0,0.0,0.0);
				if(spInputNormals.isValid())
				{
					faceNormal=
							spInputNormals->spatialVector(primitiveIndex);
				}

				if(magnitudeSquared(faceNormal)<1e-6 && subCount>2)
				{
					const SpatialVector pointA=
							spInputVertices->spatialVector(
							primitiveIndex,0);
					const SpatialVector pointB=
							spInputVertices->spatialVector(
							primitiveIndex,1);
					const SpatialVector pointC=
							spInputVertices->spatialVector(
							primitiveIndex,2);

					faceNormal=unitSafe(cross(pointC-pointA,pointB-pointA));
				}

				if(magnitudeSquared(faceNormal)<1e-6)
				{
					continue;
				}

				for(U32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const I32 pointIndexA=spInputVertices->integer(
							primitiveIndex,subIndex);
					const I32 pointIndexB=spInputVertices->integer(
							primitiveIndex,(subIndex+1)%subCount);

					const SpatialVector pointA=
							spInputPosition->spatialVector(pointIndexA);
					const SpatialVector pointB=
							spInputPosition->spatialVector(pointIndexB);
					const SpatialVector midpoint=0.5*(pointA+pointB);

					const U64 edgeKey=(pointIndexA>pointIndexB)?
							(U64(pointIndexA)<<32)+pointIndexB:
							(U64(pointIndexB)<<32)+pointIndexA;

					Edge& rEdge=edgeMap[edgeKey];

					//* not cameraDir
					const SpatialVector towardDir=
							unitSafe(midpoint-cameraPos);

					const Real towardDot=dot(towardDir,faceNormal);
					const I32 facing=(towardDot<0.0)? -1: 1;

					if(!rEdge.m_facing)
					{
						rEdge.m_facing=facing;
					}
					else if(facing != rEdge.m_facing)
					{
						//* draw

						line[0]=pointA;
						line[1]=pointB;

						spDrawBrush->drawLines(line,NULL,2,
								DrawI::e_discrete,false,&alphaBlack);
					}
				}
			}
		}

		spDrawOverlay->popDrawMode();
		spDrawBrush->popDrawMode();

		return;
	}

	const BWORD inputReplaced=
			catalogOrDefault<bool>("Input Surface","replaced",true);
	if(inputReplaced || m_spInput.isNull())
	{
		if(!access(m_spInput,"Input Surface",e_warning)) return;
	}

	if(!access(m_spInputAccessible,"Input Surface")) return;

#if FE_RUL_DEBUG
	feLog("RulerOp::handle done\n");
#endif
}

} // namespace
